@@ -16,6 +16,7 @@
 
 package io.grpc.alts.internal;
 
+import static com.google.common.base.Preconditions.checkNotNull;
 import static com.google.common.truth.Truth.assertThat;
 import static java.nio.charset.StandardCharsets.UTF_8;
 import static org.junit.Assert.assertEquals;
@@ -30,12 +31,14 @@ import io.grpc.InternalChannelz;
 import io.grpc.ManagedChannel;
 import io.grpc.SecurityLevel;
 import io.grpc.alts.internal.AltsProtocolNegotiator.LazyChannel;
+import io.grpc.alts.internal.AltsProtocolNegotiator.ServerAltsProtocolNegotiator;
 import io.grpc.alts.internal.TsiFrameProtector.Consumer;
 import io.grpc.alts.internal.TsiPeer.Property;
 import io.grpc.internal.FixedObjectPool;
 import io.grpc.internal.GrpcAttributes;
 import io.grpc.internal.ObjectPool;
 import io.grpc.netty.GrpcHttp2ConnectionHandler;
+import io.grpc.netty.InternalProtocolNegotiationEvent;
 import io.grpc.netty.NettyChannelBuilder;
 import io.netty.buffer.ByteBuf;
 import io.netty.buffer.ByteBufAllocator;
@@ -46,6 +49,7 @@ import io.netty.channel.ChannelFuture;
 import io.netty.channel.ChannelFutureListener;
 import io.netty.channel.ChannelHandler;
 import io.netty.channel.ChannelHandlerContext;
+import io.netty.channel.ChannelInboundHandlerAdapter;
 import io.netty.channel.ChannelPromise;
 import io.netty.channel.embedded.EmbeddedChannel;
 import io.netty.handler.codec.http2.DefaultHttp2Connection;
@@ -80,6 +84,7 @@ import org.junit.runners.JUnit4;
 
 /** Tests for {@link AltsProtocolNegotiator}. */
 @RunWith(JUnit4.class)
+@SuppressWarnings("FutureReturnValueIgnored")
 public class AltsProtocolNegotiatorTest {
 
   private final CapturingGrpcHttp2ConnectionHandler grpcHandler = capturingGrpcHandler();
@@ -90,9 +95,6 @@ public class AltsProtocolNegotiatorTest {
   private EmbeddedChannel channel;
   private Throwable caughtException;
 
-  private volatile TsiHandshakeHandler.TsiHandshakeCompletionEvent tsiEvent;
-  private ChannelHandler handler;
-
   private TsiPeer mockedTsiPeer = new TsiPeer(Collections.<Property<?>>emptyList());
   private AltsAuthContext mockedAltsContext =
       new AltsAuthContext(
@@ -102,12 +104,12 @@ public class AltsProtocolNegotiatorTest {
   private final TsiHandshaker mockHandshaker =
       new DelegatingTsiHandshaker(FakeTsiHandshaker.newFakeHandshakerServer()) {
         @Override
-        public TsiPeer extractPeer() throws GeneralSecurityException {
+        public TsiPeer extractPeer() {
           return mockedTsiPeer;
         }
 
         @Override
-        public Object extractPeerObject() throws GeneralSecurityException {
+        public Object extractPeerObject() {
           return mockedAltsContext;
         }
       };
@@ -115,24 +117,13 @@ public class AltsProtocolNegotiatorTest {
 
   @Before
   public void setup() throws Exception {
-    ChannelHandler userEventHandler =
-        new ChannelDuplexHandler() {
-          @Override
-          public void userEventTriggered(ChannelHandlerContext ctx, Object evt) throws Exception {
-            if (evt instanceof TsiHandshakeHandler.TsiHandshakeCompletionEvent) {
-              tsiEvent = (TsiHandshakeHandler.TsiHandshakeCompletionEvent) evt;
-            } else {
-              super.userEventTriggered(ctx, evt);
-            }
-          }
-        };
-
     ChannelHandler uncaughtExceptionHandler =
         new ChannelDuplexHandler() {
           @Override
           public void exceptionCaught(ChannelHandlerContext ctx, Throwable cause) throws Exception {
             caughtException = cause;
             super.exceptionCaught(ctx, cause);
+            ctx.close();
           }
         };
 
@@ -156,10 +147,12 @@ public class AltsProtocolNegotiatorTest {
     ManagedChannel fakeChannel = NettyChannelBuilder.forTarget("localhost:8080").build();
     ObjectPool<Channel> fakeChannelPool = new FixedObjectPool<Channel>(fakeChannel);
     LazyChannel lazyFakeChannel = new LazyChannel(fakeChannelPool);
-    handler =
-        AltsProtocolNegotiator.createServerNegotiator(handshakerFactory, lazyFakeChannel)
-            .newHandler(grpcHandler);
-    channel = new EmbeddedChannel(uncaughtExceptionHandler, handler, userEventHandler);
+    ChannelHandler altsServerHandler = new ServerAltsProtocolNegotiator(
+        handshakerFactory, lazyFakeChannel)
+        .newHandler(grpcHandler);
+    // On real server, WBAEH fires default ProtocolNegotiationEvent. KickNH provides this behavior.
+    ChannelHandler handler = new KickNegotiationHandler(altsServerHandler);
+    channel = new EmbeddedChannel(uncaughtExceptionHandler, handler);
   }
 
   @After
@@ -182,6 +175,8 @@ public class AltsProtocolNegotiatorTest {
   @Test
   @SuppressWarnings("unchecked") // List cast
   public void protectShouldRoundtrip() throws Exception {
+    doHandshake();
+
     // Write the message 1 character at a time. The message should be buffered
     // and not interfere with the handshake.
     final AtomicInteger writeCount = new AtomicInteger();
@@ -204,10 +199,6 @@ public class AltsProtocolNegotiatorTest {
     }
     channel.flush();
 
-    // Now do the handshake. The buffered message will automatically be protected
-    // and sent.
-    doHandshake();
-
     // Capture the protected data written to the wire.
     assertEquals(1, channel.outboundMessages().size());
     ByteBuf protectedData = channel.readOutbound();
@@ -351,7 +342,7 @@ public class AltsProtocolNegotiatorTest {
     doHandshake();
 
     assertThat(grpcHandler.attrs.get(AltsProtocolNegotiator.TSI_PEER_KEY)).isEqualTo(mockedTsiPeer);
-    assertThat(grpcHandler.attrs.get(AltsProtocolNegotiator.ALTS_CONTEXT_KEY))
+    assertThat(grpcHandler.attrs.get(AltsProtocolNegotiator.AUTH_CONTEXT_KEY))
         .isEqualTo(mockedAltsContext);
     assertThat(grpcHandler.attrs.get(Grpc.TRANSPORT_ATTR_REMOTE_ADDR).toString())
         .isEqualTo("embedded");
@@ -388,12 +379,12 @@ public class AltsProtocolNegotiatorTest {
     if (caughtException != null) {
       throw new RuntimeException(caughtException);
     }
-    assertNotNull(tsiEvent);
+    assertNotNull(grpcHandler.attrs);
   }
 
   private CapturingGrpcHttp2ConnectionHandler capturingGrpcHandler() {
     // Netty Boilerplate.  We don't really need any of this, but there is a tight coupling
-    // between a Http2ConnectionHandler and its dependencies.
+    // between an Http2ConnectionHandler and its dependencies.
     Http2Connection connection = new DefaultHttp2Connection(true);
     Http2FrameWriter frameWriter = new DefaultHttp2FrameWriter();
     Http2FrameReader frameReader = new DefaultHttp2FrameReader(false);
@@ -418,7 +409,8 @@ public class AltsProtocolNegotiatorTest {
 
     @Override
     public void handleProtocolNegotiationCompleted(
-        Attributes attrs, InternalChannelz.Security securityInfo) {
+        Attributes attrs,
+        @SuppressWarnings("UnusedVariable") InternalChannelz.Security securityInfo) {
       // If we are added to the pipeline, we need to remove ourselves.  The HTTP2 handler
       channel.pipeline().remove(this);
       this.attrs = attrs;
@@ -487,6 +479,11 @@ public class AltsProtocolNegotiatorTest {
       protectors.add(protector);
       return protector;
     }
+
+    @Override
+    public void close() {
+      delegate.close();
+    }
   }
 
   private static class InterceptingProtector implements TsiFrameProtector {
@@ -517,4 +514,21 @@ public class AltsProtocolNegotiatorTest {
       delegate.destroy();
     }
   }
+
+  /** Kicks off negotiation of the server. */
+  private static final class KickNegotiationHandler extends ChannelInboundHandlerAdapter {
+
+    private final ChannelHandler next;
+
+    KickNegotiationHandler(ChannelHandler next) {
+      this.next = checkNotNull(next, "next");
+    }
+
+    @Override
+    public void handlerAdded(ChannelHandlerContext ctx) throws Exception {
+      super.handlerAdded(ctx);
+      ctx.pipeline().replace(ctx.name(), /*newName= */ null, next);
+      ctx.pipeline().fireUserEventTriggered(InternalProtocolNegotiationEvent.getDefault());
+    }
+  }
 }
