@@ -28,11 +28,10 @@ import io.grpc.CompressorRegistry;
 import io.grpc.DecompressorRegistry;
 import io.grpc.EquivalentAddressGroup;
 import io.grpc.InternalChannelz;
-import io.grpc.LoadBalancer;
 import io.grpc.ManagedChannel;
 import io.grpc.ManagedChannelBuilder;
 import io.grpc.NameResolver;
-import io.grpc.NameResolverProvider;
+import io.grpc.NameResolverRegistry;
 import io.grpc.ProxyDetector;
 import io.opencensus.trace.Tracing;
 import java.net.SocketAddress;
@@ -85,9 +84,6 @@ public abstract class AbstractManagedChannelImplBuilder
   private static final ObjectPool<? extends Executor> DEFAULT_EXECUTOR_POOL =
       SharedResourcePool.forResource(GrpcUtil.SHARED_CHANNEL_EXECUTOR);
 
-  private static final NameResolver.Factory DEFAULT_NAME_RESOLVER_FACTORY =
-      NameResolverProvider.asFactory();
-
   private static final DecompressorRegistry DEFAULT_DECOMPRESSOR_REGISTRY =
       DecompressorRegistry.getDefaultInstance();
 
@@ -99,10 +95,13 @@ public abstract class AbstractManagedChannelImplBuilder
 
   ObjectPool<? extends Executor> executorPool = DEFAULT_EXECUTOR_POOL;
 
+  ObjectPool<? extends Executor> offloadExecutorPool = DEFAULT_EXECUTOR_POOL;
+
   private final List<ClientInterceptor> interceptors = new ArrayList<>();
+  final NameResolverRegistry nameResolverRegistry = NameResolverRegistry.getDefaultRegistry();
 
   // Access via getter, which may perform authority override as needed
-  private NameResolver.Factory nameResolverFactory = DEFAULT_NAME_RESOLVER_FACTORY;
+  private NameResolver.Factory nameResolverFactory = nameResolverRegistry.asFactory();
 
   final String target;
 
@@ -116,8 +115,6 @@ public abstract class AbstractManagedChannelImplBuilder
   @Nullable
   String authorityOverride;
 
-  @Nullable LoadBalancer.Factory loadBalancerFactory;
-
   String defaultLbPolicy = GrpcUtil.DEFAULT_LB_POLICY;
 
   boolean fullStreamDecompression;
@@ -222,6 +219,16 @@ public abstract class AbstractManagedChannelImplBuilder
     return thisT();
   }
 
+  @Override
+  public final T offloadExecutor(Executor executor) {
+    if (executor != null) {
+      this.offloadExecutorPool = new FixedObjectPool<>(executor);
+    } else {
+      this.offloadExecutorPool = DEFAULT_EXECUTOR_POOL;
+    }
+    return thisT();
+  }
+
   @Override
   public final T intercept(List<ClientInterceptor> interceptors) {
     this.interceptors.addAll(interceptors);
@@ -241,21 +248,11 @@ public abstract class AbstractManagedChannelImplBuilder
     if (resolverFactory != null) {
       this.nameResolverFactory = resolverFactory;
     } else {
-      this.nameResolverFactory = DEFAULT_NAME_RESOLVER_FACTORY;
+      this.nameResolverFactory = nameResolverRegistry.asFactory();
     }
     return thisT();
   }
 
-  @Deprecated
-  @Override
-  public final T loadBalancerFactory(LoadBalancer.Factory loadBalancerFactory) {
-    Preconditions.checkState(directServerAddress == null,
-        "directServerAddress is set (%s), which forbids the use of LoadBalancer.Factory",
-        directServerAddress);
-    this.loadBalancerFactory = loadBalancerFactory;
-    return thisT();
-  }
-
   @Override
   public final T defaultLoadBalancingPolicy(String policy) {
     Preconditions.checkState(directServerAddress == null,
@@ -595,7 +592,7 @@ public abstract class AbstractManagedChannelImplBuilder
     }
 
     @Override
-    public NameResolver newNameResolver(URI notUsedUri, NameResolver.Helper helper) {
+    public NameResolver newNameResolver(URI notUsedUri, NameResolver.Args args) {
       return new NameResolver() {
         @Override
         public String getServiceAuthority() {
@@ -603,10 +600,12 @@ public abstract class AbstractManagedChannelImplBuilder
         }
 
         @Override
-        public void start(final Listener listener) {
-          listener.onAddresses(
-              Collections.singletonList(new EquivalentAddressGroup(address)),
-              Attributes.EMPTY);
+        public void start(Listener2 listener) {
+          listener.onResult(
+              ResolutionResult.newBuilder()
+                  .setAddresses(Collections.singletonList(new EquivalentAddressGroup(address)))
+                  .setAttributes(Attributes.EMPTY)
+                  .build());
         }
 
         @Override
