@@ -46,6 +46,9 @@ import io.grpc.Metadata;
 import io.grpc.MethodDescriptor;
 import io.grpc.MethodDescriptor.MethodType;
 import io.grpc.Status;
+import io.perfmark.Link;
+import io.perfmark.PerfMark;
+import io.perfmark.Tag;
 import java.io.InputStream;
 import java.nio.charset.Charset;
 import java.util.concurrent.CancellationException;
@@ -67,6 +70,7 @@ final class ClientCallImpl<ReqT, RespT> extends ClientCall<ReqT, RespT> {
       = "gzip".getBytes(Charset.forName("US-ASCII"));
 
   private final MethodDescriptor<ReqT, RespT> method;
+  private final Tag tag;
   private final Executor callExecutor;
   private final CallTracer channelCallsTracer;
   private final Context context;
@@ -92,6 +96,8 @@ final class ClientCallImpl<ReqT, RespT> extends ClientCall<ReqT, RespT> {
       CallTracer channelCallsTracer,
       boolean retryEnabled) {
     this.method = method;
+    // TODO(carl-mastrangelo): consider moving this construction to ManagedChannelImpl.
+    this.tag = PerfMark.createTag(method.getFullMethodName(), System.identityHashCode(this));
     // If we know that the executor is a direct executor, we don't need to wrap it with a
     // SerializingExecutor. This is purely for performance reasons.
     // See https://github.com/grpc/grpc-java/issues/368
@@ -107,6 +113,7 @@ final class ClientCallImpl<ReqT, RespT> extends ClientCall<ReqT, RespT> {
     this.clientTransportProvider = clientTransportProvider;
     this.deadlineCancellationExecutor = deadlineCancellationExecutor;
     this.retryEnabled = retryEnabled;
+    PerfMark.event("ClientCall.<init>", tag);
   }
 
   private final class ContextCancellationListener implements CancellationListener {
@@ -177,7 +184,16 @@ final class ClientCallImpl<ReqT, RespT> extends ClientCall<ReqT, RespT> {
   }
 
   @Override
-  public void start(final Listener<RespT> observer, Metadata headers) {
+  public void start(Listener<RespT> observer, Metadata headers) {
+    PerfMark.startTask("ClientCall.start", tag);
+    try {
+      startInternal(observer, headers);
+    } finally {
+      PerfMark.stopTask("ClientCall.start", tag);
+    }
+  }
+
+  private void startInternal(final Listener<RespT> observer, Metadata headers) {
     checkState(stream == null, "Already started");
     checkState(!cancelCalled, "call was cancelled");
     checkNotNull(observer, "observer");
@@ -234,7 +250,7 @@ final class ClientCallImpl<ReqT, RespT> extends ClientCall<ReqT, RespT> {
     boolean deadlineExceeded = effectiveDeadline != null && effectiveDeadline.isExpired();
     if (!deadlineExceeded) {
       logIfContextNarrowedTimeout(
-          effectiveDeadline, callOptions.getDeadline(), context.getDeadline());
+          effectiveDeadline, context.getDeadline(), callOptions.getDeadline());
       if (retryEnabled) {
         stream = clientTransportProvider.newRetriableStream(method, callOptions, headers, context);
       } else {
@@ -249,7 +265,8 @@ final class ClientCallImpl<ReqT, RespT> extends ClientCall<ReqT, RespT> {
       }
     } else {
       stream = new FailingClientStream(
-          DEADLINE_EXCEEDED.withDescription("deadline exceeded: " + effectiveDeadline));
+          DEADLINE_EXCEEDED.withDescription(
+              "ClientCall started after deadline exceeded: " + effectiveDeadline));
     }
 
     if (callOptions.getAuthority() != null) {
@@ -279,7 +296,7 @@ final class ClientCallImpl<ReqT, RespT> extends ClientCall<ReqT, RespT> {
     context.addListener(cancellationListener, directExecutor());
     if (effectiveDeadline != null
         // If the context has the effective deadline, we don't need to schedule an extra task.
-        && context.getDeadline() != effectiveDeadline
+        && !effectiveDeadline.equals(context.getDeadline())
         // If the channel has been terminated, we don't need to schedule an extra task.
         && deadlineCancellationExecutor != null) {
       deadlineCancellationFuture = startDeadlineTimer(effectiveDeadline);
@@ -297,7 +314,7 @@ final class ClientCallImpl<ReqT, RespT> extends ClientCall<ReqT, RespT> {
       Deadline effectiveDeadline, @Nullable Deadline outerCallDeadline,
       @Nullable Deadline callDeadline) {
     if (!log.isLoggable(Level.FINE) || effectiveDeadline == null
-        || outerCallDeadline != effectiveDeadline) {
+        || !effectiveDeadline.equals(outerCallDeadline)) {
       return;
     }
 
@@ -331,10 +348,23 @@ final class ClientCallImpl<ReqT, RespT> extends ClientCall<ReqT, RespT> {
 
     @Override
     public void run() {
+      InsightBuilder insight = new InsightBuilder();
+      stream.appendTimeoutInsight(insight);
       // DelayedStream.cancel() is safe to call from a thread that is different from where the
       // stream is created.
-      stream.cancel(DEADLINE_EXCEEDED.augmentDescription(
-          String.format("deadline exceeded after %dns", remainingNanos)));
+      long seconds = Math.abs(remainingNanos) / TimeUnit.SECONDS.toNanos(1);
+      long nanos = Math.abs(remainingNanos) % TimeUnit.SECONDS.toNanos(1);
+
+      StringBuilder buf = new StringBuilder();
+      buf.append("deadline exceeded after ");
+      if (remainingNanos < 0) {
+        buf.append('-');
+      }
+      buf.append(seconds);
+      buf.append(String.format(".%09d", nanos));
+      buf.append("s. ");
+      buf.append(insight);
+      stream.cancel(DEADLINE_EXCEEDED.augmentDescription(buf.toString()));
     }
   }
 
@@ -364,13 +394,27 @@ final class ClientCallImpl<ReqT, RespT> extends ClientCall<ReqT, RespT> {
 
   @Override
   public void request(int numMessages) {
-    checkState(stream != null, "Not started");
-    checkArgument(numMessages >= 0, "Number requested must be non-negative");
-    stream.request(numMessages);
+    PerfMark.startTask("ClientCall.request", tag);
+    try {
+      checkState(stream != null, "Not started");
+      checkArgument(numMessages >= 0, "Number requested must be non-negative");
+      stream.request(numMessages);
+    } finally {
+      PerfMark.stopTask("ClientCall.cancel", tag);
+    }
   }
 
   @Override
   public void cancel(@Nullable String message, @Nullable Throwable cause) {
+    PerfMark.startTask("ClientCall.cancel", tag);
+    try {
+      cancelInternal(message, cause);
+    } finally {
+      PerfMark.stopTask("ClientCall.cancel", tag);
+    }
+  }
+
+  private void cancelInternal(@Nullable String message, @Nullable Throwable cause) {
     if (message == null && cause == null) {
       cause = new CancellationException("Cancelled without a message or cause");
       log.log(Level.WARNING, "Cancelling without a message or cause is suboptimal", cause);
@@ -401,6 +445,15 @@ final class ClientCallImpl<ReqT, RespT> extends ClientCall<ReqT, RespT> {
 
   @Override
   public void halfClose() {
+    PerfMark.startTask("ClientCall.halfClose", tag);
+    try {
+      halfCloseInternal();
+    } finally {
+      PerfMark.stopTask("ClientCall.halfClose", tag);
+    }
+  }
+
+  private void halfCloseInternal() {
     checkState(stream != null, "Not started");
     checkState(!cancelCalled, "call was cancelled");
     checkState(!halfCloseCalled, "call already half-closed");
@@ -410,6 +463,15 @@ final class ClientCallImpl<ReqT, RespT> extends ClientCall<ReqT, RespT> {
 
   @Override
   public void sendMessage(ReqT message) {
+    PerfMark.startTask("ClientCall.sendMessage", tag);
+    try {
+      sendMessageInternal(message);
+    } finally {
+      PerfMark.stopTask("ClientCall.sendMessage", tag);
+    }
+  }
+
+  private void sendMessageInternal(ReqT message) {
     checkState(stream != null, "Not started");
     checkState(!cancelCalled, "call was cancelled");
     checkState(!halfCloseCalled, "call was half-closed");
@@ -474,17 +536,30 @@ final class ClientCallImpl<ReqT, RespT> extends ClientCall<ReqT, RespT> {
 
     @Override
     public void headersRead(final Metadata headers) {
-      class HeadersRead extends ContextRunnable {
+      PerfMark.startTask("ClientStreamListener.headersRead", tag);
+      final Link link = PerfMark.linkOut();
+
+      final class HeadersRead extends ContextRunnable {
         HeadersRead() {
           super(context);
         }
 
         @Override
-        public final void runInContext() {
+        public void runInContext() {
+          PerfMark.startTask("ClientCall$Listener.headersRead", tag);
+          PerfMark.linkIn(link);
+          try {
+            runInternal();
+          } finally {
+            PerfMark.stopTask("ClientCall$Listener.headersRead", tag);
+          }
+        }
+
+        private void runInternal() {
+          if (closed) {
+            return;
+          }
           try {
-            if (closed) {
-              return;
-            }
             observer.onHeaders(headers);
           } catch (Throwable t) {
             Status status =
@@ -495,25 +570,41 @@ final class ClientCallImpl<ReqT, RespT> extends ClientCall<ReqT, RespT> {
         }
       }
 
-      callExecutor.execute(new HeadersRead());
+      try {
+        callExecutor.execute(new HeadersRead());
+      } finally {
+        PerfMark.stopTask("ClientStreamListener.headersRead", tag);
+      }
     }
 
     @Override
     public void messagesAvailable(final MessageProducer producer) {
-      class MessagesAvailable extends ContextRunnable {
+      PerfMark.startTask("ClientStreamListener.messagesAvailable", tag);
+      final Link link = PerfMark.linkOut();
+
+      final class MessagesAvailable extends ContextRunnable {
         MessagesAvailable() {
           super(context);
         }
 
         @Override
-        public final void runInContext() {
+        public void runInContext() {
+          PerfMark.startTask("ClientCall$Listener.messagesAvailable", tag);
+          PerfMark.linkIn(link);
+          try {
+            runInternal();
+          } finally {
+            PerfMark.stopTask("ClientCall$Listener.messagesAvailable", tag);
+          }
+        }
+
+        private void runInternal() {
           if (closed) {
             GrpcUtil.closeQuietly(producer);
             return;
           }
-
-          InputStream message;
           try {
+            InputStream message;
             while ((message = producer.next()) != null) {
               try {
                 observer.onMessage(method.parseResponse(message));
@@ -533,7 +624,11 @@ final class ClientCallImpl<ReqT, RespT> extends ClientCall<ReqT, RespT> {
         }
       }
 
-      callExecutor.execute(new MessagesAvailable());
+      try {
+        callExecutor.execute(new MessagesAvailable());
+      } finally {
+        PerfMark.stopTask("ClientStreamListener.messagesAvailable", tag);
+      }
     }
 
     /**
@@ -557,26 +652,50 @@ final class ClientCallImpl<ReqT, RespT> extends ClientCall<ReqT, RespT> {
 
     @Override
     public void closed(Status status, RpcProgress rpcProgress, Metadata trailers) {
+      PerfMark.startTask("ClientStreamListener.closed", tag);
+      try {
+        closedInternal(status, rpcProgress, trailers);
+      } finally {
+        PerfMark.stopTask("ClientStreamListener.closed", tag);
+      }
+    }
+
+    private void closedInternal(
+        Status status, @SuppressWarnings("unused") RpcProgress rpcProgress, Metadata trailers) {
       Deadline deadline = effectiveDeadline();
       if (status.getCode() == Status.Code.CANCELLED && deadline != null) {
         // When the server's deadline expires, it can only reset the stream with CANCEL and no
         // description. Since our timer may be delayed in firing, we double-check the deadline and
         // turn the failure into the likely more helpful DEADLINE_EXCEEDED status.
         if (deadline.isExpired()) {
-          status = DEADLINE_EXCEEDED;
+          InsightBuilder insight = new InsightBuilder();
+          stream.appendTimeoutInsight(insight);
+          status = DEADLINE_EXCEEDED.augmentDescription(
+              "ClientCall was cancelled at or after deadline. " + insight);
           // Replace trailers to prevent mixing sources of status and trailers.
           trailers = new Metadata();
         }
       }
       final Status savedStatus = status;
       final Metadata savedTrailers = trailers;
-      class StreamClosed extends ContextRunnable {
+      final Link link = PerfMark.linkOut();
+      final class StreamClosed extends ContextRunnable {
         StreamClosed() {
           super(context);
         }
 
         @Override
-        public final void runInContext() {
+        public void runInContext() {
+          PerfMark.startTask("ClientCall$Listener.onClose", tag);
+          PerfMark.linkIn(link);
+          try {
+            runInternal();
+          } finally {
+            PerfMark.stopTask("ClientCall$Listener.onClose", tag);
+          }
+        }
+
+        private void runInternal() {
           if (closed) {
             // We intentionally don't keep the status or metadata from the server.
             return;
@@ -590,13 +709,30 @@ final class ClientCallImpl<ReqT, RespT> extends ClientCall<ReqT, RespT> {
 
     @Override
     public void onReady() {
-      class StreamOnReady extends ContextRunnable {
+      if (method.getType().clientSendsOneMessage()) {
+        return;
+      }
+
+      PerfMark.startTask("ClientStreamListener.onReady", tag);
+      final Link link = PerfMark.linkOut();
+
+      final class StreamOnReady extends ContextRunnable {
         StreamOnReady() {
           super(context);
         }
 
         @Override
-        public final void runInContext() {
+        public void runInContext() {
+          PerfMark.startTask("ClientCall$Listener.onReady", tag);
+          PerfMark.linkIn(link);
+          try {
+            runInternal();
+          } finally {
+            PerfMark.stopTask("ClientCall$Listener.onReady", tag);
+          }
+        }
+
+        private void runInternal() {
           try {
             observer.onReady();
           } catch (Throwable t) {
@@ -608,7 +744,11 @@ final class ClientCallImpl<ReqT, RespT> extends ClientCall<ReqT, RespT> {
         }
       }
 
-      callExecutor.execute(new StreamOnReady());
+      try {
+        callExecutor.execute(new StreamOnReady());
+      } finally {
+        PerfMark.stopTask("ClientStreamListener.onReady", tag);
+      }
     }
   }
 }
