@@ -33,7 +33,6 @@ import io.grpc.ClientCall;
 import io.grpc.ClientInterceptor;
 import io.grpc.CompressorRegistry;
 import io.grpc.DecompressorRegistry;
-import io.grpc.LoadBalancer;
 import io.grpc.MethodDescriptor;
 import io.grpc.NameResolver;
 import io.grpc.internal.testing.StatsTestUtils.FakeStatsRecorder;
@@ -100,6 +99,21 @@ public class AbstractManagedChannelImplBuilderTest {
     assertEquals(MoreExecutors.directExecutor(), builder.executorPool.getObject());
   }
 
+  @Test
+  public void offloadExecutor_normal() {
+    Executor executor = mock(Executor.class);
+    assertEquals(builder, builder.offloadExecutor(executor));
+    assertEquals(executor, builder.offloadExecutorPool.getObject());
+  }
+
+  @Test
+  public void offloadExecutor_null() {
+    ObjectPool<? extends Executor> defaultValue = builder.offloadExecutorPool;
+    builder.offloadExecutor(mock(Executor.class));
+    assertEquals(builder, builder.offloadExecutor(null));
+    assertEquals(defaultValue, builder.offloadExecutorPool);
+  }
+
   @Test
   public void nameResolverFactory_default() {
     assertNotNull(builder.getNameResolverFactory());
@@ -125,34 +139,6 @@ public class AbstractManagedChannelImplBuilderTest {
     directAddressBuilder.nameResolverFactory(mock(NameResolver.Factory.class));
   }
 
-  @Test
-  public void loadBalancerFactory_default() {
-    assertNull(builder.loadBalancerFactory);
-  }
-
-  @Test
-  @Deprecated
-  public void loadBalancerFactory_normal() {
-    LoadBalancer.Factory loadBalancerFactory = mock(LoadBalancer.Factory.class);
-    assertEquals(builder, builder.loadBalancerFactory(loadBalancerFactory));
-    assertEquals(loadBalancerFactory, builder.loadBalancerFactory);
-  }
-
-  @Test
-  @Deprecated
-  public void loadBalancerFactory_null() {
-    LoadBalancer.Factory defaultValue = builder.loadBalancerFactory;
-    builder.loadBalancerFactory(mock(LoadBalancer.Factory.class));
-    assertEquals(builder, builder.loadBalancerFactory(null));
-    assertEquals(defaultValue, builder.loadBalancerFactory);
-  }
-
-  @Test(expected = IllegalStateException.class)
-  @Deprecated
-  public void loadBalancerFactory_notAllowedWithDirectAddress() {
-    directAddressBuilder.loadBalancerFactory(mock(LoadBalancer.Factory.class));
-  }
-
   @Test
   public void defaultLoadBalancingPolicy_default() {
     assertEquals("pick_first", builder.defaultLbPolicy);
@@ -301,7 +287,7 @@ public class AbstractManagedChannelImplBuilderTest {
         .isInstanceOf(CensusTracingModule.TracingClientInterceptor.class);
     assertThat(effectiveInterceptors.get(1))
         .isInstanceOf(CensusStatsModule.StatsClientInterceptor.class);
-    assertThat(effectiveInterceptors.get(2)).isSameAs(DUMMY_USER_INTERCEPTOR);
+    assertThat(effectiveInterceptors.get(2)).isSameInstanceAs(DUMMY_USER_INTERCEPTOR);
   }
 
   @Test
@@ -312,7 +298,7 @@ public class AbstractManagedChannelImplBuilderTest {
     assertEquals(2, effectiveInterceptors.size());
     assertThat(effectiveInterceptors.get(0))
         .isInstanceOf(CensusTracingModule.TracingClientInterceptor.class);
-    assertThat(effectiveInterceptors.get(1)).isSameAs(DUMMY_USER_INTERCEPTOR);
+    assertThat(effectiveInterceptors.get(1)).isSameInstanceAs(DUMMY_USER_INTERCEPTOR);
   }
 
   @Test
@@ -323,7 +309,7 @@ public class AbstractManagedChannelImplBuilderTest {
     assertEquals(2, effectiveInterceptors.size());
     assertThat(effectiveInterceptors.get(0))
         .isInstanceOf(CensusStatsModule.StatsClientInterceptor.class);
-    assertThat(effectiveInterceptors.get(1)).isSameAs(DUMMY_USER_INTERCEPTOR);
+    assertThat(effectiveInterceptors.get(1)).isSameInstanceAs(DUMMY_USER_INTERCEPTOR);
   }
 
   @Test
