@@ -34,6 +34,7 @@ import static org.mockito.Mockito.verifyNoMoreInteractions;
 import static org.mockito.Mockito.verifyZeroInteractions;
 import static org.mockito.Mockito.when;
 
+import com.google.common.base.Preconditions;
 import io.grpc.Attributes;
 import io.grpc.ChannelLogger;
 import io.grpc.ChannelLogger.ChannelLogLevel;
@@ -41,14 +42,16 @@ import io.grpc.ConnectivityState;
 import io.grpc.ConnectivityStateInfo;
 import io.grpc.EquivalentAddressGroup;
 import io.grpc.LoadBalancer;
+import io.grpc.LoadBalancer.CreateSubchannelArgs;
 import io.grpc.LoadBalancer.Helper;
+import io.grpc.LoadBalancer.ResolvedAddresses;
 import io.grpc.LoadBalancer.Subchannel;
 import io.grpc.LoadBalancer.SubchannelPicker;
+import io.grpc.LoadBalancer.SubchannelStateListener;
 import io.grpc.LoadBalancerProvider;
 import io.grpc.LoadBalancerRegistry;
 import io.grpc.ManagedChannel;
 import io.grpc.Status;
-import io.grpc.Status.Code;
 import io.grpc.SynchronizationContext;
 import io.grpc.grpclb.GrpclbLoadBalancerProvider;
 import io.grpc.internal.AutoConfiguredLoadBalancerFactory.AutoConfiguredLoadBalancer;
@@ -109,15 +112,14 @@ public class AutoConfiguredLoadBalancerFactoryTest {
 
   @Test
   public void newLoadBalancer_isAuto() {
-    LoadBalancer lb = lbf.newLoadBalancer(new TestHelper());
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(new TestHelper());
 
     assertThat(lb).isInstanceOf(AutoConfiguredLoadBalancer.class);
   }
 
   @Test
   public void defaultIsPickFirst() {
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) lbf.newLoadBalancer(new TestHelper());
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(new TestHelper());
 
     assertThat(lb.getDelegateProvider()).isInstanceOf(PickFirstLoadBalancerProvider.class);
     assertThat(lb.getDelegate().getClass().getName()).contains("PickFirst");
@@ -125,18 +127,17 @@ public class AutoConfiguredLoadBalancerFactoryTest {
 
   @Test
   public void defaultIsConfigurable() {
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) new AutoConfiguredLoadBalancerFactory("test_lb")
+    AutoConfiguredLoadBalancer lb = new AutoConfiguredLoadBalancerFactory("test_lb")
         .newLoadBalancer(new TestHelper());
 
-    assertThat(lb.getDelegateProvider()).isSameAs(testLbBalancerProvider);
-    assertThat(lb.getDelegate()).isSameAs(testLbBalancer);
+    assertThat(lb.getDelegateProvider()).isSameInstanceAs(testLbBalancerProvider);
+    assertThat(lb.getDelegate()).isSameInstanceAs(testLbBalancer);
   }
 
+  @SuppressWarnings("deprecation")
   @Test
   public void forwardsCalls() {
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) lbf.newLoadBalancer(new TestHelper());
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(new TestHelper());
 
     final AtomicInteger calls = new AtomicInteger();
     TestLoadBalancer testlb = new TestLoadBalancer() {
@@ -175,18 +176,22 @@ public class AutoConfiguredLoadBalancerFactoryTest {
         Collections.singletonList(new EquivalentAddressGroup(new SocketAddress(){}));
     Helper helper = new TestHelper() {
       @Override
-      public Subchannel createSubchannel(List<EquivalentAddressGroup> addrs, Attributes attrs) {
-        assertThat(addrs).isEqualTo(servers);
-        return new TestSubchannel(addrs, attrs);
+      public Subchannel createSubchannel(CreateSubchannelArgs args) {
+        assertThat(args.getAddresses()).isEqualTo(servers);
+        return new TestSubchannel(args);
       }
     };
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) lbf.newLoadBalancer(helper);
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(helper);
     LoadBalancer oldDelegate = lb.getDelegate();
 
-    lb.handleResolvedAddressGroups(servers, Attributes.EMPTY);
+    Status handleResult = lb.tryHandleResolvedAddresses(
+        ResolvedAddresses.newBuilder()
+            .setAddresses(servers)
+            .setAttributes(Attributes.EMPTY)
+            .build());
 
-    assertThat(lb.getDelegate()).isSameAs(oldDelegate);
+    assertThat(handleResult.getCode()).isEqualTo(Status.Code.OK);
+    assertThat(lb.getDelegate()).isSameInstanceAs(oldDelegate);
   }
 
   @Test
@@ -201,13 +206,12 @@ public class AutoConfiguredLoadBalancerFactoryTest {
         Collections.singletonList(new EquivalentAddressGroup(new SocketAddress(){}));
     Helper helper = new TestHelper() {
       @Override
-      public Subchannel createSubchannel(List<EquivalentAddressGroup> addrs, Attributes attrs) {
-        assertThat(addrs).isEqualTo(servers);
-        return new TestSubchannel(addrs, attrs);
+      public Subchannel createSubchannel(CreateSubchannelArgs args) {
+        assertThat(args.getAddresses()).isEqualTo(servers);
+        return new TestSubchannel(args);
       }
     };
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) lbf.newLoadBalancer(helper);
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(helper);
     final AtomicBoolean shutdown = new AtomicBoolean();
     TestLoadBalancer testlb = new TestLoadBalancer() {
 
@@ -216,11 +220,6 @@ public class AutoConfiguredLoadBalancerFactoryTest {
         // noop
       }
 
-      @Override
-      public void handleSubchannelState(Subchannel subchannel, ConnectivityStateInfo stateInfo) {
-        // noop
-      }
-
       @Override
       public void shutdown() {
         shutdown.set(true);
@@ -228,8 +227,13 @@ public class AutoConfiguredLoadBalancerFactoryTest {
     };
     lb.setDelegate(testlb);
 
-    lb.handleResolvedAddressGroups(servers, serviceConfigAttrs);
+    Status handleResult = lb.tryHandleResolvedAddresses(
+        ResolvedAddresses.newBuilder()
+            .setAddresses(servers)
+            .setAttributes(serviceConfigAttrs)
+            .build());
 
+    assertThat(handleResult.getCode()).isEqualTo(Status.Code.OK);
     assertThat(lb.getDelegateProvider().getClass().getName()).isEqualTo(
         "io.grpc.util.SecretRoundRobinLoadBalancerProvider$Provider");
     assertTrue(shutdown.get());
@@ -246,16 +250,23 @@ public class AutoConfiguredLoadBalancerFactoryTest {
     final List<EquivalentAddressGroup> servers =
         Collections.singletonList(new EquivalentAddressGroup(new SocketAddress(){}));
     Helper helper = new TestHelper();
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) lbf.newLoadBalancer(helper);
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(helper);
 
-    lb.handleResolvedAddressGroups(servers, serviceConfigAttrs);
+    Status handleResult = lb.tryHandleResolvedAddresses(
+        ResolvedAddresses.newBuilder()
+            .setAddresses(servers)
+            .setAttributes(serviceConfigAttrs)
+            .build());
 
     verify(testLbBalancerProvider).newLoadBalancer(same(helper));
-    assertThat(lb.getDelegate()).isSameAs(testLbBalancer);
-    ArgumentCaptor<Attributes> attrsCaptor = ArgumentCaptor.forClass(null);
-    verify(testLbBalancer).handleResolvedAddressGroups(eq(servers), attrsCaptor.capture());
-    assertThat(attrsCaptor.getValue().get(ATTR_LOAD_BALANCING_CONFIG))
+    assertThat(handleResult.getCode()).isEqualTo(Status.Code.OK);
+    assertThat(lb.getDelegate()).isSameInstanceAs(testLbBalancer);
+    ArgumentCaptor<ResolvedAddresses> resultCaptor =
+        ArgumentCaptor.forClass(ResolvedAddresses.class);
+    verify(testLbBalancer).handleResolvedAddresses(resultCaptor.capture());
+    assertThat(resultCaptor.getValue().getAddresses()).containsExactlyElementsIn(servers).inOrder();
+    Attributes actualAttributes = resultCaptor.getValue().getAttributes();
+    assertThat(actualAttributes.get(ATTR_LOAD_BALANCING_CONFIG))
         .isEqualTo(Collections.singletonMap("setting1", "high"));
     verify(testLbBalancer, atLeast(0)).canHandleEmptyAddressListFromNameResolution();
     verifyNoMoreInteractions(testLbBalancer);
@@ -266,12 +277,20 @@ public class AutoConfiguredLoadBalancerFactoryTest {
         Attributes.newBuilder()
             .set(GrpcAttributes.NAME_RESOLVER_SERVICE_CONFIG, serviceConfig)
             .build();
-    lb.handleResolvedAddressGroups(servers, serviceConfigAttrs);
+    handleResult = lb.tryHandleResolvedAddresses(
+        ResolvedAddresses.newBuilder()
+            .setAddresses(servers)
+            .setAttributes(serviceConfigAttrs)
+            .build());
 
-    verify(testLbBalancer, times(2))
-        .handleResolvedAddressGroups(eq(servers), attrsCaptor.capture());
+    resultCaptor =
+        ArgumentCaptor.forClass(ResolvedAddresses.class);
+    verify(testLbBalancer, times(2)).handleResolvedAddresses(resultCaptor.capture());
+    assertThat(handleResult.getCode()).isEqualTo(Status.Code.OK);
+    assertThat(resultCaptor.getValue().getAddresses()).containsExactlyElementsIn(servers).inOrder();
+    actualAttributes = resultCaptor.getValue().getAttributes();
     // But the balancer config is changed.
-    assertThat(attrsCaptor.getValue().get(ATTR_LOAD_BALANCING_CONFIG))
+    assertThat(actualAttributes.get(ATTR_LOAD_BALANCING_CONFIG))
         .isEqualTo(Collections.singletonMap("setting1", "low"));
     // Service config didn't change policy, thus the delegateLb is not swapped
     verifyNoMoreInteractions(testLbBalancer);
@@ -292,71 +311,79 @@ public class AutoConfiguredLoadBalancerFactoryTest {
                 new SocketAddress(){},
                 Attributes.newBuilder().set(GrpcAttributes.ATTR_LB_ADDR_AUTHORITY, "ok").build()));
     Helper helper = new TestHelper();
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) new AutoConfiguredLoadBalancerFactory(
+    AutoConfiguredLoadBalancer lb = new AutoConfiguredLoadBalancerFactory(
             registry, GrpcUtil.DEFAULT_LB_POLICY).newLoadBalancer(helper);
 
-    lb.handleResolvedAddressGroups(servers, Attributes.EMPTY);
+    Status handleResult = lb.tryHandleResolvedAddresses(
+        ResolvedAddresses.newBuilder()
+            .setAddresses(servers)
+            .setAttributes(Attributes.EMPTY)
+            .build());
 
-    assertThat(lb.getDelegate()).isSameAs(testLbBalancer);
-    verify(testLbBalancer).handleResolvedAddressGroups(
-        eq(Collections.singletonList(servers.get(0))), any(Attributes.class));
+    assertThat(handleResult.getCode()).isEqualTo(Status.Code.OK);
+    assertThat(lb.getDelegate()).isSameInstanceAs(testLbBalancer);
+    verify(testLbBalancer).handleResolvedAddresses(
+        ResolvedAddresses.newBuilder()
+            .setAddresses(Collections.singletonList(servers.get(0)))
+            .setAttributes(Attributes.EMPTY)
+            .build());
   }
 
   @Test
   public void handleResolvedAddressGroups_delegateDoNotAcceptEmptyAddressList_nothing()
       throws Exception {
     Helper helper = new TestHelper();
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) lbf.newLoadBalancer(helper);
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(helper);
 
     Map<String, ?> serviceConfig =
         parseConfig("{\"loadBalancingConfig\": [ {\"test_lb\": { \"setting1\": \"high\" } } ] }");
-    lb.handleResolvedAddressGroups(
-        Collections.<EquivalentAddressGroup>emptyList(),
-        Attributes.newBuilder()
-            .set(GrpcAttributes.NAME_RESOLVER_SERVICE_CONFIG, serviceConfig).build());
+    Status handleResult = lb.tryHandleResolvedAddresses(
+        ResolvedAddresses.newBuilder()
+            .setAddresses(Collections.<EquivalentAddressGroup>emptyList())
+            .setAttributes(Attributes.newBuilder()
+                .set(GrpcAttributes.NAME_RESOLVER_SERVICE_CONFIG, serviceConfig).build())
+            .build());
 
-    assertThat(lb.getDelegate()).isSameAs(testLbBalancer);
     assertThat(testLbBalancer.canHandleEmptyAddressListFromNameResolution()).isFalse();
-    ArgumentCaptor<Status> statusCaptor = ArgumentCaptor.forClass(null);
-    verify(testLbBalancer).handleNameResolutionError(statusCaptor.capture());
-    Status error = statusCaptor.getValue();
-    assertThat(error.getCode()).isEqualTo(Code.UNAVAILABLE);
-    assertThat(error.getDescription()).startsWith("Name resolver returned no usable address");
+    assertThat(handleResult.getCode()).isEqualTo(Status.Code.UNAVAILABLE);
+    assertThat(handleResult.getDescription()).startsWith("NameResolver returned no usable address");
+    assertThat(lb.getDelegate()).isSameInstanceAs(testLbBalancer);
   }
 
   @Test
   public void handleResolvedAddressGroups_delegateAcceptsEmptyAddressList()
       throws Exception {
     Helper helper = new TestHelper();
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) lbf.newLoadBalancer(helper);
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(helper);
 
     Map<String, ?> serviceConfig =
         parseConfig("{\"loadBalancingConfig\": [ {\"test_lb2\": { \"setting1\": \"high\" } } ] }");
-    lb.handleResolvedAddressGroups(
-        Collections.<EquivalentAddressGroup>emptyList(),
-        Attributes.newBuilder()
-            .set(GrpcAttributes.NAME_RESOLVER_SERVICE_CONFIG, serviceConfig).build());
+    Status handleResult = lb.tryHandleResolvedAddresses(
+        ResolvedAddresses.newBuilder()
+            .setAddresses(Collections.<EquivalentAddressGroup>emptyList())
+            .setAttributes(Attributes.newBuilder()
+                .set(GrpcAttributes.NAME_RESOLVER_SERVICE_CONFIG, serviceConfig).build())
+            .build());
 
-    assertThat(lb.getDelegate()).isSameAs(testLbBalancer2);
+    assertThat(handleResult.getCode()).isEqualTo(Status.Code.OK);
+    assertThat(lb.getDelegate()).isSameInstanceAs(testLbBalancer2);
     assertThat(testLbBalancer2.canHandleEmptyAddressListFromNameResolution()).isTrue();
-    ArgumentCaptor<Attributes> attrsCaptor = ArgumentCaptor.forClass(null);
-    verify(testLbBalancer2).handleResolvedAddressGroups(
-        eq(Collections.<EquivalentAddressGroup>emptyList()), attrsCaptor.capture());
-    Map<String, ?> lbConfig =
-        attrsCaptor.getValue().get(LoadBalancer.ATTR_LOAD_BALANCING_CONFIG);
+    ArgumentCaptor<ResolvedAddresses> resultCaptor =
+        ArgumentCaptor.forClass(ResolvedAddresses.class);
+    verify(testLbBalancer2).handleResolvedAddresses(resultCaptor.capture());
+    assertThat(resultCaptor.getValue().getAddresses()).isEmpty();
+    Attributes actualAttributes = resultCaptor.getValue().getAttributes();
+
+    Map<String, ?> lbConfig = actualAttributes.get(LoadBalancer.ATTR_LOAD_BALANCING_CONFIG);
     assertThat(lbConfig).isEqualTo(Collections.<String, Object>singletonMap("setting1", "high"));
-    assertThat(attrsCaptor.getValue().get(GrpcAttributes.NAME_RESOLVER_SERVICE_CONFIG))
-        .isSameAs(serviceConfig);
+    assertThat(actualAttributes.get(GrpcAttributes.NAME_RESOLVER_SERVICE_CONFIG))
+        .isSameInstanceAs(serviceConfig);
   }
 
   @Test
   public void decideLoadBalancerProvider_noBalancerAddresses_noServiceConfig_pickFirst()
       throws Exception {
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) lbf.newLoadBalancer(new TestHelper());
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(new TestHelper());
     Map<String, ?> serviceConfig = null;
     List<EquivalentAddressGroup> servers =
         Collections.singletonList(new EquivalentAddressGroup(new SocketAddress(){}));
@@ -371,15 +398,14 @@ public class AutoConfiguredLoadBalancerFactoryTest {
   @Test
   public void decideLoadBalancerProvider_noBalancerAddresses_noServiceConfig_customDefault()
       throws Exception {
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) new AutoConfiguredLoadBalancerFactory("test_lb")
-            .newLoadBalancer(new TestHelper());
+    AutoConfiguredLoadBalancer lb = new AutoConfiguredLoadBalancerFactory("test_lb")
+        .newLoadBalancer(new TestHelper());
     Map<String, ?> serviceConfig = null;
     List<EquivalentAddressGroup> servers =
         Collections.singletonList(new EquivalentAddressGroup(new SocketAddress(){}));
     PolicySelection selection = lb.decideLoadBalancerProvider(servers, serviceConfig);
 
-    assertThat(selection.provider).isSameAs(testLbBalancerProvider);
+    assertThat(selection.provider).isSameInstanceAs(testLbBalancerProvider);
     assertThat(selection.serverList).isEqualTo(servers);
     assertThat(selection.config).isNull();
     verifyZeroInteractions(channelLogger);
@@ -387,8 +413,7 @@ public class AutoConfiguredLoadBalancerFactoryTest {
 
   @Test
   public void decideLoadBalancerProvider_oneBalancer_noServiceConfig_grpclb() throws Exception {
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) lbf.newLoadBalancer(new TestHelper());
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(new TestHelper());
     Map<String, ?> serviceConfig = null;
     List<EquivalentAddressGroup> servers =
         Collections.singletonList(
@@ -405,8 +430,7 @@ public class AutoConfiguredLoadBalancerFactoryTest {
 
   @Test
   public void decideLoadBalancerProvider_serviceConfigLbPolicy() throws Exception {
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) lbf.newLoadBalancer(new TestHelper());
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(new TestHelper());
     Map<String, String> serviceConfig = new HashMap<>();
     serviceConfig.put("loadBalancingPolicy", "round_robin");
     List<EquivalentAddressGroup> servers =
@@ -429,8 +453,7 @@ public class AutoConfiguredLoadBalancerFactoryTest {
   @SuppressWarnings("unchecked")
   @Test
   public void decideLoadBalancerProvider_serviceConfigLbConfig() throws Exception {
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) lbf.newLoadBalancer(new TestHelper());
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(new TestHelper());
     Map<String, ?> serviceConfig =
         parseConfig("{\"loadBalancingConfig\": [ {\"round_robin\": {} } ] }");
     List<EquivalentAddressGroup> servers =
@@ -452,8 +475,7 @@ public class AutoConfiguredLoadBalancerFactoryTest {
 
   @Test
   public void decideLoadBalancerProvider_grpclbConfigPropagated() throws Exception {
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) lbf.newLoadBalancer(new TestHelper());
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(new TestHelper());
     Map<String, ?> serviceConfig =
         parseConfig(
             "{\"loadBalancingConfig\": ["
@@ -475,8 +497,7 @@ public class AutoConfiguredLoadBalancerFactoryTest {
 
   @Test
   public void decideLoadBalancerProvider_policyUnavailButGrpclbAddressPresent() throws Exception {
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) lbf.newLoadBalancer(new TestHelper());
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(new TestHelper());
     Map<String, ?> serviceConfig =
         parseConfig(
             "{\"loadBalancingConfig\": ["
@@ -503,9 +524,8 @@ public class AutoConfiguredLoadBalancerFactoryTest {
     LoadBalancerProvider fakeRoundRobinProvider =
         new FakeLoadBalancerProvider("round_robin", testLbBalancer);
     registry.register(fakeRoundRobinProvider);
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) new AutoConfiguredLoadBalancerFactory(
-            registry, GrpcUtil.DEFAULT_LB_POLICY).newLoadBalancer(new TestHelper());
+    AutoConfiguredLoadBalancer lb = new AutoConfiguredLoadBalancerFactory(
+        registry, GrpcUtil.DEFAULT_LB_POLICY).newLoadBalancer(new TestHelper());
     Map<String, ?> serviceConfig =
         parseConfig("{\"loadBalancingConfig\": [ {\"grpclb\": {} } ] }");
     List<EquivalentAddressGroup> servers =
@@ -516,7 +536,7 @@ public class AutoConfiguredLoadBalancerFactoryTest {
             new EquivalentAddressGroup(new SocketAddress(){}));
     PolicySelection selection = lb.decideLoadBalancerProvider(servers, serviceConfig);
 
-    assertThat(selection.provider).isSameAs(fakeRoundRobinProvider);
+    assertThat(selection.provider).isSameInstanceAs(fakeRoundRobinProvider);
     assertThat(selection.config).isNull();
     verify(channelLogger).log(
         eq(ChannelLogLevel.ERROR),
@@ -525,7 +545,7 @@ public class AutoConfiguredLoadBalancerFactoryTest {
     // Called for the second time, the warning is only logged once
     selection = lb.decideLoadBalancerProvider(servers, serviceConfig);
 
-    assertThat(selection.provider).isSameAs(fakeRoundRobinProvider);
+    assertThat(selection.provider).isSameInstanceAs(fakeRoundRobinProvider);
     // Balancer addresses are filtered out in the server list passed to round_robin
     assertThat(selection.serverList).containsExactly(servers.get(1));
     assertThat(selection.config).isNull();
@@ -538,9 +558,8 @@ public class AutoConfiguredLoadBalancerFactoryTest {
     LoadBalancerRegistry registry = new LoadBalancerRegistry();
     registry.register(new PickFirstLoadBalancerProvider());
     registry.register(new FakeLoadBalancerProvider("round_robin", testLbBalancer));
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) new AutoConfiguredLoadBalancerFactory(
-            registry, GrpcUtil.DEFAULT_LB_POLICY).newLoadBalancer(new TestHelper());
+    AutoConfiguredLoadBalancer lb = new AutoConfiguredLoadBalancerFactory(
+        registry, GrpcUtil.DEFAULT_LB_POLICY).newLoadBalancer(new TestHelper());
     Map<String, ?> serviceConfig =
         parseConfig("{\"loadBalancingConfig\": [ {\"grpclb\": {} } ] }");
     List<EquivalentAddressGroup> servers =
@@ -560,8 +579,7 @@ public class AutoConfiguredLoadBalancerFactoryTest {
 
   @Test
   public void decideLoadBalancerProvider_serviceConfigLbPolicyOverridesDefault() throws Exception {
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) lbf.newLoadBalancer(new TestHelper());
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(new TestHelper());
     Map<String, String> serviceConfig = new HashMap<>();
     serviceConfig.put("loadBalancingPolicy", "round_robin");
     List<EquivalentAddressGroup> servers =
@@ -576,8 +594,7 @@ public class AutoConfiguredLoadBalancerFactoryTest {
 
   @Test
   public void decideLoadBalancerProvider_serviceConfigLbConfigOverridesDefault() throws Exception {
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) lbf.newLoadBalancer(new TestHelper());
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(new TestHelper());
     Map<String, ?> serviceConfig =
         parseConfig("{\"loadBalancingConfig\": [ {\"round_robin\": {\"setting1\": \"high\"} } ] }");
     List<EquivalentAddressGroup> servers =
@@ -593,8 +610,7 @@ public class AutoConfiguredLoadBalancerFactoryTest {
 
   @Test
   public void decideLoadBalancerProvider_serviceConfigLbPolicyFailsOnUnknown() {
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) lbf.newLoadBalancer(new TestHelper());
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(new TestHelper());
     Map<String, String> serviceConfig = new HashMap<>();
     serviceConfig.put("loadBalancingPolicy", "MAGIC_BALANCER");
     List<EquivalentAddressGroup> servers =
@@ -610,8 +626,7 @@ public class AutoConfiguredLoadBalancerFactoryTest {
 
   @Test
   public void decideLoadBalancerProvider_serviceConfigLbConfigFailsOnUnknown() throws Exception {
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) lbf.newLoadBalancer(new TestHelper());
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(new TestHelper());
     Map<String, ?> serviceConfig =
         parseConfig("{\"loadBalancingConfig\": [ {\"magic_balancer\": {} } ] }");
     List<EquivalentAddressGroup> servers =
@@ -627,8 +642,7 @@ public class AutoConfiguredLoadBalancerFactoryTest {
 
   @Test
   public void decideLoadBalancerProvider_serviceConfigLbConfigSkipUnknown() throws Exception {
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) lbf.newLoadBalancer(new TestHelper());
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(new TestHelper());
     Map<String, ?> serviceConfig =
         parseConfig(
             "{\"loadBalancingConfig\": [ {\"magic_balancer\": {} }, {\"round_robin\": {} } ] }");
@@ -648,8 +662,7 @@ public class AutoConfiguredLoadBalancerFactoryTest {
 
   @Test
   public void decideLoadBalancerProvider_serviceConfigHasZeroLbConfig() throws Exception {
-    AutoConfiguredLoadBalancer lb =
-        (AutoConfiguredLoadBalancer) lbf.newLoadBalancer(new TestHelper());
+    AutoConfiguredLoadBalancer lb = lbf.newLoadBalancer(new TestHelper());
     List<EquivalentAddressGroup> servers =
         Collections.singletonList(new EquivalentAddressGroup(new SocketAddress(){}));
     PolicySelection selection = lb.decideLoadBalancerProvider(
@@ -668,8 +681,17 @@ public class AutoConfiguredLoadBalancerFactoryTest {
         Collections.singletonList(new EquivalentAddressGroup(new SocketAddress(){}));
     Helper helper = new TestHelper() {
       @Override
+      @Deprecated
       public Subchannel createSubchannel(List<EquivalentAddressGroup> addrs, Attributes attrs) {
-        return new TestSubchannel(addrs, attrs);
+        return new TestSubchannel(CreateSubchannelArgs.newBuilder()
+            .setAddresses(addrs)
+            .setAttributes(attrs)
+            .build());
+      }
+
+      @Override
+      public Subchannel createSubchannel(CreateSubchannelArgs args) {
+        return new TestSubchannel(args);
       }
 
       @Override
@@ -699,18 +721,27 @@ public class AutoConfiguredLoadBalancerFactoryTest {
       }
     };
 
-    LoadBalancer lb = new AutoConfiguredLoadBalancerFactory(GrpcUtil.DEFAULT_LB_POLICY)
-        .newLoadBalancer(helper);
-    lb.handleResolvedAddressGroups(servers, Attributes.EMPTY);
+    AutoConfiguredLoadBalancer lb =
+        new AutoConfiguredLoadBalancerFactory(GrpcUtil.DEFAULT_LB_POLICY).newLoadBalancer(helper);
+    Status handleResult = lb.tryHandleResolvedAddresses(
+        ResolvedAddresses.newBuilder()
+            .setAddresses(servers)
+            .setAttributes(Attributes.EMPTY)
+            .build());
 
+    assertThat(handleResult.getCode()).isEqualTo(Status.Code.OK);
     verifyNoMoreInteractions(channelLogger);
 
     Map<String, String> serviceConfig = new HashMap<>();
     serviceConfig.put("loadBalancingPolicy", "round_robin");
-    lb.handleResolvedAddressGroups(servers,
-        Attributes.newBuilder()
-            .set(GrpcAttributes.NAME_RESOLVER_SERVICE_CONFIG, serviceConfig).build());
+    handleResult = lb.tryHandleResolvedAddresses(
+        ResolvedAddresses.newBuilder()
+            .setAddresses(servers)
+            .setAttributes(Attributes.newBuilder()
+                .set(GrpcAttributes.NAME_RESOLVER_SERVICE_CONFIG, serviceConfig).build())
+            .build());
 
+    assertThat(handleResult.getCode()).isEqualTo(Status.Code.OK);
     verify(channelLogger).log(
         eq(ChannelLogLevel.INFO),
         eq("Load balancer changed from {0} to {1}"),
@@ -722,9 +753,13 @@ public class AutoConfiguredLoadBalancerFactoryTest {
     verifyNoMoreInteractions(channelLogger);
 
     serviceConfig.put("loadBalancingPolicy", "round_robin");
-    lb.handleResolvedAddressGroups(servers,
-        Attributes.newBuilder()
-            .set(GrpcAttributes.NAME_RESOLVER_SERVICE_CONFIG, serviceConfig).build());
+    handleResult = lb.tryHandleResolvedAddresses(
+        ResolvedAddresses.newBuilder()
+            .setAddresses(servers)
+            .setAttributes(Attributes.newBuilder()
+                .set(GrpcAttributes.NAME_RESOLVER_SERVICE_CONFIG, serviceConfig).build())
+            .build());
+    assertThat(handleResult.getCode()).isEqualTo(Status.Code.OK);
     verify(channelLogger, times(2)).log(
         eq(ChannelLogLevel.DEBUG),
         eq("Load-balancing config: {0}"),
@@ -734,8 +769,13 @@ public class AutoConfiguredLoadBalancerFactoryTest {
     servers = Collections.singletonList(new EquivalentAddressGroup(
         new SocketAddress(){},
         Attributes.newBuilder().set(GrpcAttributes.ATTR_LB_ADDR_AUTHORITY, "ok").build()));
-    lb.handleResolvedAddressGroups(servers, Attributes.EMPTY);
+    handleResult = lb.tryHandleResolvedAddresses(
+        ResolvedAddresses.newBuilder()
+            .setAddresses(servers)
+            .setAttributes(Attributes.EMPTY)
+            .build());
 
+    assertThat(handleResult.getCode()).isEqualTo(Status.Code.OK);
     verify(channelLogger).log(
         eq(ChannelLogLevel.INFO),
         eq("Load balancer changed from {0} to {1}"),
@@ -756,19 +796,20 @@ public class AutoConfiguredLoadBalancerFactoryTest {
     }
 
     @Override
+    @Deprecated
     public void handleResolvedAddressGroups(
         List<EquivalentAddressGroup> servers, Attributes attributes) {
       delegate().handleResolvedAddressGroups(servers, attributes);
     }
 
     @Override
-    public void handleNameResolutionError(Status error) {
-      delegate().handleNameResolutionError(error);
+    public void handleResolvedAddresses(ResolvedAddresses resolvedAddresses) {
+      delegate().handleResolvedAddresses(resolvedAddresses);
     }
 
     @Override
-    public void handleSubchannelState(Subchannel subchannel, ConnectivityStateInfo stateInfo) {
-      delegate().handleSubchannelState(subchannel, stateInfo);
+    public void handleNameResolutionError(Status error) {
+      delegate().handleNameResolutionError(error);
     }
 
     @Override
@@ -806,14 +847,18 @@ public class AutoConfiguredLoadBalancerFactoryTest {
   }
 
   private static class TestSubchannel extends Subchannel {
-    TestSubchannel(List<EquivalentAddressGroup> addrs, Attributes attrs) {
-      this.addrs = addrs;
-      this.attrs = attrs;
+    TestSubchannel(CreateSubchannelArgs args) {
+      this.addrs = args.getAddresses();
+      this.attrs = args.getAttributes();
     }
 
-    final List<EquivalentAddressGroup> addrs;
+    List<EquivalentAddressGroup> addrs;
     final Attributes attrs;
 
+    @Override
+    public void start(SubchannelStateListener listener) {
+    }
+
     @Override
     public void shutdown() {
     }
@@ -831,6 +876,12 @@ public class AutoConfiguredLoadBalancerFactoryTest {
     public Attributes getAttributes() {
       return attrs;
     }
+
+    @Override
+    public void updateAddresses(List<EquivalentAddressGroup> addrs) {
+      Preconditions.checkNotNull(addrs, "addrs");
+      this.addrs = addrs;
+    }
   }
 
   private static final class FakeLoadBalancerProvider extends LoadBalancerProvider {
