@@ -28,6 +28,7 @@ import static org.junit.Assert.fail;
 import static org.mockito.ArgumentMatchers.any;
 import static org.mockito.ArgumentMatchers.eq;
 import static org.mockito.ArgumentMatchers.same;
+import static org.mockito.Mockito.doAnswer;
 import static org.mockito.Mockito.doThrow;
 import static org.mockito.Mockito.never;
 import static org.mockito.Mockito.timeout;
@@ -78,8 +79,10 @@ import org.mockito.ArgumentCaptor;
 import org.mockito.ArgumentMatchers;
 import org.mockito.Captor;
 import org.mockito.Mock;
+import org.mockito.invocation.InvocationOnMock;
 import org.mockito.junit.MockitoJUnit;
 import org.mockito.junit.MockitoRule;
+import org.mockito.stubbing.Answer;
 
 /**
  * Test for {@link ClientCallImpl}.
@@ -136,6 +139,14 @@ public class ClientCallImplTest {
     when(transport.newStream(
             any(MethodDescriptor.class), any(Metadata.class), any(CallOptions.class)))
         .thenReturn(stream);
+    doAnswer(new Answer<Void>() {
+        @Override
+        public Void answer(InvocationOnMock in) {
+          InsightBuilder insight = (InsightBuilder) in.getArguments()[0];
+          insight.appendKeyValue("remote_addr", "127.0.0.1:443");
+          return null;
+        }
+      }).when(stream).appendTimeoutInsight(any(InsightBuilder.class));
     baseCallOptions = CallOptions.DEFAULT.withStreamTracerFactory(streamTracerFactory);
   }
 
@@ -200,7 +211,7 @@ public class ClientCallImplTest {
         ArgumentMatchers.isA(Metadata.class));
     Status callListenerStatus = statusArgumentCaptor.getValue();
     assertThat(callListenerStatus.getCode()).isEqualTo(Status.Code.CANCELLED);
-    assertThat(callListenerStatus.getCause()).isSameAs(failure);
+    assertThat(callListenerStatus.getCause()).isSameInstanceAs(failure);
     verify(stream).cancel(same(callListenerStatus));
   }
 
@@ -236,7 +247,7 @@ public class ClientCallImplTest {
         ArgumentMatchers.isA(Metadata.class));
     Status callListenerStatus = statusArgumentCaptor.getValue();
     assertThat(callListenerStatus.getCode()).isEqualTo(Status.Code.CANCELLED);
-    assertThat(callListenerStatus.getCause()).isSameAs(failure);
+    assertThat(callListenerStatus.getCause()).isSameInstanceAs(failure);
     verify(stream).cancel(same(callListenerStatus));
   }
 
@@ -244,7 +255,7 @@ public class ClientCallImplTest {
   public void exceptionInOnReadyTakesPrecedenceOverServer() {
     DelayedExecutor executor = new DelayedExecutor();
     ClientCallImpl<Void, Void> call = new ClientCallImpl<>(
-        method,
+        method.toBuilder().setType(MethodType.UNKNOWN).build(),
         executor,
         baseCallOptions,
         provider,
@@ -272,7 +283,7 @@ public class ClientCallImplTest {
         ArgumentMatchers.isA(Metadata.class));
     Status callListenerStatus = statusArgumentCaptor.getValue();
     assertThat(callListenerStatus.getCode()).isEqualTo(Status.Code.CANCELLED);
-    assertThat(callListenerStatus.getCause()).isSameAs(failure);
+    assertThat(callListenerStatus.getCause()).isSameInstanceAs(failure);
     verify(stream).cancel(same(callListenerStatus));
   }
 
@@ -510,7 +521,7 @@ public class ClientCallImplTest {
     Context previous = context.attach();
 
     ClientCallImpl<Void, Void> call = new ClientCallImpl<>(
-        method,
+        method.toBuilder().setType(MethodType.UNKNOWN).build(),
         new SerializingExecutor(Executors.newSingleThreadExecutor()),
         baseCallOptions,
         provider,
@@ -676,6 +687,8 @@ public class ClientCallImplTest {
         .newStream(any(MethodDescriptor.class), any(Metadata.class), any(CallOptions.class));
     verify(callListener, timeout(1000)).onClose(statusCaptor.capture(), any(Metadata.class));
     assertEquals(Status.Code.DEADLINE_EXCEEDED, statusCaptor.getValue().getCode());
+    assertThat(statusCaptor.getValue().getDescription())
+        .startsWith("ClientCall started after deadline exceeded");
     verifyZeroInteractions(provider);
   }
 
@@ -807,6 +820,8 @@ public class ClientCallImplTest {
 
     verify(stream, times(1)).cancel(statusCaptor.capture());
     assertEquals(Status.Code.DEADLINE_EXCEEDED, statusCaptor.getValue().getCode());
+    assertThat(statusCaptor.getValue().getDescription())
+        .matches("deadline exceeded after [0-9]+\\.[0-9]+s. \\[remote_addr=127\\.0\\.0\\.1:443\\]");
   }
 
   @Test
@@ -834,6 +849,7 @@ public class ClientCallImplTest {
 
     verify(stream, times(1)).cancel(statusCaptor.capture());
     assertEquals(Status.Code.DEADLINE_EXCEEDED, statusCaptor.getValue().getCode());
+    assertThat(statusCaptor.getValue().getDescription()).isEqualTo("context timed out");
   }
 
   @Test
