@@ -71,6 +71,7 @@ import java.util.Map;
 import java.util.Map.Entry;
 import java.util.concurrent.ScheduledExecutorService;
 import java.util.concurrent.TimeUnit;
+import java.util.concurrent.atomic.AtomicBoolean;
 import java.util.concurrent.atomic.AtomicReference;
 import javax.annotation.Nullable;
 import javax.annotation.concurrent.NotThreadSafe;
@@ -103,11 +104,13 @@ final class GrpclbState {
       }
     };
 
-  static enum Mode {
+  enum Mode {
     ROUND_ROBIN,
     PICK_FIRST,
   }
 
+  // backend list for pick_first will be picked from this index
+  private final int pickFirstIndex;
   private final String serviceName;
   private final Helper helper;
   private final SynchronizationContext syncContext;
@@ -157,7 +160,8 @@ final class GrpclbState {
       SubchannelPool subchannelPool,
       TimeProvider time,
       Stopwatch stopwatch,
-      BackoffPolicy.Provider backoffPolicyProvider) {
+      BackoffPolicy.Provider backoffPolicyProvider,
+      int pickFirstIndex) {
     this.mode = checkNotNull(mode, "mode");
     this.helper = checkNotNull(helper, "helper");
     this.syncContext = checkNotNull(helper.getSynchronizationContext(), "syncContext");
@@ -169,6 +173,7 @@ final class GrpclbState {
     this.backoffPolicyProvider = checkNotNull(backoffPolicyProvider, "backoffPolicyProvider");
     this.serviceName = checkNotNull(helper.getAuthority(), "helper returns null authority");
     this.logger = checkNotNull(helper.getChannelLogger(), "logger");
+    this.pickFirstIndex = pickFirstIndex;
   }
 
   void handleSubchannelState(Subchannel subchannel, ConnectivityStateInfo newState) {
@@ -224,6 +229,14 @@ final class GrpclbState {
     maybeUpdatePicker();
   }
 
+  void requestConnection() {
+    for (RoundRobinEntry entry : currentPicker.pickList) {
+      if (entry instanceof IdleSubchannelEntry) {
+        ((IdleSubchannelEntry) entry).subchannel.requestConnection();
+      }
+    }
+  }
+
   private void maybeUseFallbackBackends() {
     if (balancerWorking) {
       return;
@@ -266,7 +279,7 @@ final class GrpclbState {
 
   private void shutdownLbRpc() {
     if (lbStream != null) {
-      lbStream.close(null);
+      lbStream.close(Status.CANCELLED.withDescription("balancer shutdown").asException());
       // lbStream will be set to null in LbStream.cleanup()
     }
   }
@@ -328,8 +341,10 @@ final class GrpclbState {
         subchannelPool.clear();
         break;
       case PICK_FIRST:
-        checkState(subchannels.size() == 1, "Excessive Subchannels: %s", subchannels);
-        subchannels.values().iterator().next().shutdown();
+        if (!subchannels.isEmpty()) {
+          checkState(subchannels.size() == 1, "Excessive Subchannels: %s", subchannels);
+          subchannels.values().iterator().next().shutdown();
+        }
         break;
       default:
         throw new AssertionError("Missing case for " + mode);
@@ -363,6 +378,7 @@ final class GrpclbState {
   /**
    * Populate the round-robin lists with the given values.
    */
+  @SuppressWarnings("deprecation")
   private void useRoundRobinLists(
       List<DropEntry> newDropList, List<BackendAddressGroup> newBackendAddrList,
       @Nullable GrpclbClientLoadRecorder loadRecorder) {
@@ -405,12 +421,14 @@ final class GrpclbState {
         subchannels = Collections.unmodifiableMap(newSubchannelMap);
         break;
       case PICK_FIRST:
-        List<EquivalentAddressGroup> eagList = new ArrayList<>();
+        int numOfEags = newBackendAddrList.size();
+        List<EquivalentAddressGroup> eagList = Arrays.asList(new EquivalentAddressGroup[numOfEags]);
         // Because for PICK_FIRST, we create a single Subchannel for all addresses, we have to
         // attach the tokens to the EAG attributes and use TokenAttachingLoadRecorder to put them on
         // headers.
         //
         // The PICK_FIRST code path doesn't cache Subchannels.
+        int offset = pickFirstIndex;
         for (BackendAddressGroup bag : newBackendAddrList) {
           EquivalentAddressGroup origEag = bag.getAddresses();
           Attributes eagAttrs = origEag.getAttributes();
@@ -418,15 +436,19 @@ final class GrpclbState {
             eagAttrs = eagAttrs.toBuilder()
                 .set(GrpclbConstants.TOKEN_ATTRIBUTE_KEY, bag.getToken()).build();
           }
-          eagList.add(new EquivalentAddressGroup(origEag.getAddresses(), eagAttrs));
+          eagList.set(
+              offset++ % numOfEags,
+              new EquivalentAddressGroup(origEag.getAddresses(), eagAttrs));
         }
         Subchannel subchannel;
         if (subchannels.isEmpty()) {
+          // TODO(zhangkun83): remove the deprecation suppression on this method once migrated to
+          // the new createSubchannel().
           subchannel = helper.createSubchannel(eagList, createSubchannelAttrs());
         } else {
           checkState(subchannels.size() == 1, "Unexpected Subchannel count: %s", subchannels);
           subchannel = subchannels.values().iterator().next();
-          helper.updateSubchannelAddresses(subchannel, eagList);
+          subchannel.updateAddresses(eagList);
         }
         subchannels = Collections.singletonMap(eagList, subchannel);
         newBackendList.add(
@@ -646,17 +668,13 @@ final class GrpclbState {
       helper.refreshNameResolution();
     }
 
-    void close(@Nullable Exception error) {
+    void close(Exception error) {
       if (closed) {
         return;
       }
       closed = true;
       cleanUp();
-      if (error == null) {
-        lbRequestWriter.onCompleted();
-      } else {
-        lbRequestWriter.onError(error);
-      }
+      lbRequestWriter.onError(error);
     }
 
     private void cleanUp() {
@@ -730,7 +748,7 @@ final class GrpclbState {
               break;
             default:
               pickList = Collections.<RoundRobinEntry>singletonList(
-                  new IdleSubchannelEntry(onlyEntry.subchannel));
+                  new IdleSubchannelEntry(onlyEntry.subchannel, syncContext));
           }
         }
         break;
@@ -913,15 +931,25 @@ final class GrpclbState {
 
   @VisibleForTesting
   static final class IdleSubchannelEntry implements RoundRobinEntry {
+    private final SynchronizationContext syncContext;
     private final Subchannel subchannel;
+    private final AtomicBoolean connectionRequested = new AtomicBoolean(false);
 
-    IdleSubchannelEntry(Subchannel subchannel) {
+    IdleSubchannelEntry(Subchannel subchannel, SynchronizationContext syncContext) {
       this.subchannel = checkNotNull(subchannel, "subchannel");
+      this.syncContext = checkNotNull(syncContext, "syncContext");
     }
 
     @Override
     public PickResult picked(Metadata headers) {
-      subchannel.requestConnection();
+      if (connectionRequested.compareAndSet(false, true)) {
+        syncContext.execute(new Runnable() {
+            @Override
+            public void run() {
+              subchannel.requestConnection();
+            }
+          });
+      }
       return PickResult.withNoResult();
     }
 
@@ -933,7 +961,7 @@ final class GrpclbState {
 
     @Override
     public int hashCode() {
-      return Objects.hashCode(subchannel);
+      return Objects.hashCode(subchannel, syncContext);
     }
 
     @Override
@@ -942,7 +970,8 @@ final class GrpclbState {
         return false;
       }
       IdleSubchannelEntry that = (IdleSubchannelEntry) other;
-      return Objects.equal(subchannel, that.subchannel);
+      return Objects.equal(subchannel, that.subchannel)
+          && Objects.equal(syncContext, that.syncContext);
     }
   }
 
@@ -1025,13 +1054,5 @@ final class GrpclbState {
       }
     }
 
-    @Override
-    public void requestConnection() {
-      for (RoundRobinEntry entry : pickList) {
-        if (entry instanceof IdleSubchannelEntry) {
-          ((IdleSubchannelEntry) entry).subchannel.requestConnection();
-        }
-      }
-    }
   }
 }
