@@ -22,8 +22,8 @@ import static io.grpc.internal.ClientStreamListener.RpcProgress.PROCESSED;
 import static io.grpc.internal.ClientStreamListener.RpcProgress.REFUSED;
 import static io.grpc.internal.GrpcUtil.DEFAULT_MAX_MESSAGE_SIZE;
 import static io.grpc.okhttp.Headers.CONTENT_TYPE_HEADER;
+import static io.grpc.okhttp.Headers.HTTP_SCHEME_HEADER;
 import static io.grpc.okhttp.Headers.METHOD_HEADER;
-import static io.grpc.okhttp.Headers.SCHEME_HEADER;
 import static io.grpc.okhttp.Headers.TE_HEADER;
 import static org.junit.Assert.assertEquals;
 import static org.junit.Assert.assertFalse;
@@ -74,6 +74,7 @@ import io.grpc.internal.GrpcUtil;
 import io.grpc.internal.ManagedClientTransport;
 import io.grpc.internal.TransportTracer;
 import io.grpc.okhttp.OkHttpClientTransport.ClientFrameHandler;
+import io.grpc.okhttp.OkHttpFrameLogger.Direction;
 import io.grpc.okhttp.internal.ConnectionSpec;
 import io.grpc.okhttp.internal.framed.ErrorCode;
 import io.grpc.okhttp.internal.framed.FrameReader;
@@ -104,6 +105,10 @@ import java.util.concurrent.Executors;
 import java.util.concurrent.LinkedBlockingQueue;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicBoolean;
+import java.util.logging.Handler;
+import java.util.logging.Level;
+import java.util.logging.LogRecord;
+import java.util.logging.Logger;
 import javax.annotation.Nullable;
 import javax.net.SocketFactory;
 import javax.net.ssl.HostnameVerifier;
@@ -140,6 +145,7 @@ public class OkHttpClientTransportTest {
   private static final int DEFAULT_START_STREAM_ID = 3;
   private static final int DEFAULT_MAX_INBOUND_METADATA_SIZE = Integer.MAX_VALUE;
   private static final Attributes EAG_ATTRS = Attributes.EMPTY;
+  private static final Logger logger = Logger.getLogger(OkHttpClientTransport.class.getName());
 
   @Rule public final Timeout globalTimeout = Timeout.seconds(10);
 
@@ -224,6 +230,7 @@ public class OkHttpClientTransportTest {
         executor,
         frameReader,
         frameWriter,
+        new OkHttpFrameLogger(Level.FINE, logger),
         startId,
         socket,
         stopwatchSupplier,
@@ -257,12 +264,122 @@ public class OkHttpClientTransportTest {
         NO_PROXY,
         tooManyPingsRunnable,
         DEFAULT_MAX_INBOUND_METADATA_SIZE,
-        transportTracer);
+        transportTracer,
+        false);
     String s = clientTransport.toString();
     assertTrue("Unexpected: " + s, s.contains("OkHttpClientTransport"));
     assertTrue("Unexpected: " + s, s.contains(address.toString()));
   }
 
+  /**
+   * Test logging is functioning correctly for client received Http/2 frames. Not intended to test
+   * actual frame content being logged.
+   */
+  @Test
+  public void testClientHandlerFrameLogger() throws Exception {
+    final List<LogRecord> logs = new ArrayList<>();
+    Handler handler = new Handler() {
+      @Override
+      public void publish(LogRecord record) {
+        logs.add(record);
+      }
+
+      @Override
+      public void flush() {
+      }
+
+      @Override
+      public void close() throws SecurityException {
+      }
+    };
+    logger.addHandler(handler);
+    logger.setLevel(Level.ALL);
+
+    initTransport();
+
+    MockStreamListener listener = new MockStreamListener();
+    OkHttpClientStream stream =
+        clientTransport.newStream(method, new Metadata(), CallOptions.DEFAULT);
+    stream.start(listener);
+    stream.request(1);
+
+    frameHandler().headers(false, false, 3, 0, grpcResponseHeaders(), HeadersMode.HTTP_20_HEADERS);
+    assertThat(logs).hasSize(1);
+    LogRecord log = logs.remove(0);
+    assertThat(log.getMessage()).startsWith(Direction.INBOUND + " HEADERS: streamId=" + 3);
+    assertThat(log.getLevel()).isEqualTo(Level.FINE);
+
+    final String message = "Hello Client";
+    Buffer buffer = createMessageFrame(message);
+    frameHandler().data(false, 3, buffer, (int) buffer.size());
+    assertThat(logs).hasSize(1);
+    log = logs.remove(0);
+    assertThat(log.getMessage()).startsWith(Direction.INBOUND + " DATA: streamId=" + 3);
+    assertThat(log.getLevel()).isEqualTo(Level.FINE);
+
+    // At most 64 bytes of data frame will be logged.
+    frameHandler().data(false, 3, createMessageFrame(new String(new char[1000])), 1000);
+    assertThat(logs).hasSize(1);
+    log = logs.remove(0);
+    String data = log.getMessage();
+    assertThat(data).endsWith("...");
+    assertThat(data.substring(data.indexOf("bytes="), data.indexOf("..."))).hasLength(64 * 2 + 6);
+
+    // A SETTINGS ACK frame is sent out after receiving SETTINGS frame.
+    frameHandler().settings(false, new Settings());
+    assertThat(logs).hasSize(2);
+    log = logs.remove(0);
+    assertThat(log.getMessage()).startsWith(Direction.INBOUND + " SETTINGS: ack=false");
+    assertThat(log.getLevel()).isEqualTo(Level.FINE);
+    log = logs.remove(0);
+    assertThat(log.getMessage()).startsWith(Direction.OUTBOUND + " SETTINGS: ack=true");
+    assertThat(log.getLevel()).isEqualTo(Level.FINE);
+
+    // A PING ACK frame is sent out after receiving PING frame.
+    frameHandler().ping(false, 0, 0);
+    assertThat(logs).hasSize(2);
+    log = logs.remove(0);
+    assertThat(log.getMessage()).startsWith(Direction.INBOUND + " PING: ack=false");
+    assertThat(log.getLevel()).isEqualTo(Level.FINE);
+    log = logs.remove(0);
+    assertThat(log.getMessage()).startsWith(Direction.OUTBOUND + " PING: ack=true");
+    assertThat(log.getLevel()).isEqualTo(Level.FINE);
+
+    // As server push is not supported, a RST_STREAM is sent out after receiving PUSH_PROMISE frame.
+    frameHandler().pushPromise(3, 3, grpcResponseHeaders());
+    assertThat(logs).hasSize(2);
+    log = logs.remove(0);
+    assertThat(log.getMessage()).startsWith(Direction.INBOUND + " PUSH_PROMISE");
+    assertThat(log.getLevel()).isEqualTo(Level.FINE);
+    log = logs.remove(0);
+    assertThat(log.getMessage()).startsWith(Direction.OUTBOUND + " RST_STREAM");
+    assertThat(log.getLevel()).isEqualTo(Level.FINE);
+
+    frameHandler().rstStream(3, ErrorCode.CANCEL);
+    assertThat(logs).hasSize(1);
+    log = logs.remove(0);
+    assertThat(log.getMessage()).startsWith(Direction.INBOUND + " RST_STREAM");
+    assertThat(log.getLevel()).isEqualTo(Level.FINE);
+
+    // Outbound GO_AWAY is responded after receiving inbound GO_AWAY frame.
+    frameHandler().goAway(3, ErrorCode.CANCEL, ByteString.EMPTY);
+    assertThat(logs).hasSize(2);
+    log = logs.remove(0);
+    assertThat(log.getMessage()).startsWith(Direction.INBOUND + " GO_AWAY");
+    assertThat(log.getLevel()).isEqualTo(Level.FINE);
+    log = logs.remove(0);
+    assertThat(log.getMessage()).startsWith(Direction.OUTBOUND + " GO_AWAY");
+    assertThat(log.getLevel()).isEqualTo(Level.FINE);
+
+    frameHandler().windowUpdate(3, 32);
+    assertThat(logs).hasSize(1);
+    log = logs.remove(0);
+    assertThat(log.getMessage()).startsWith(Direction.INBOUND + " WINDOW_UPDATE");
+    assertThat(log.getLevel()).isEqualTo(Level.FINE);
+
+    logger.removeHandler(handler);
+  }
+
   @Test
   public void maxMessageSizeShouldBeEnforced() throws Exception {
     // Allow the response payloads of up to 1 byte.
@@ -533,7 +650,7 @@ public class OkHttpClientTransportTest {
     stream.start(listener);
     Header userAgentHeader = new Header(GrpcUtil.USER_AGENT_KEY.name(),
             GrpcUtil.getGrpcUserAgent("okhttp", null));
-    List<Header> expectedHeaders = Arrays.asList(SCHEME_HEADER, METHOD_HEADER,
+    List<Header> expectedHeaders = Arrays.asList(HTTP_SCHEME_HEADER, METHOD_HEADER,
             new Header(Header.TARGET_AUTHORITY, "notarealauthority:80"),
             new Header(Header.TARGET_PATH, "/" + method.getFullMethodName()),
             userAgentHeader, CONTENT_TYPE_HEADER, TE_HEADER);
@@ -550,7 +667,7 @@ public class OkHttpClientTransportTest {
     OkHttpClientStream stream =
         clientTransport.newStream(method, new Metadata(), CallOptions.DEFAULT);
     stream.start(listener);
-    List<Header> expectedHeaders = Arrays.asList(SCHEME_HEADER, METHOD_HEADER,
+    List<Header> expectedHeaders = Arrays.asList(HTTP_SCHEME_HEADER, METHOD_HEADER,
         new Header(Header.TARGET_AUTHORITY, "notarealauthority:80"),
         new Header(Header.TARGET_PATH, "/" + method.getFullMethodName()),
         new Header(GrpcUtil.USER_AGENT_KEY.name(),
@@ -1548,7 +1665,8 @@ public class OkHttpClientTransportTest {
         NO_PROXY,
         tooManyPingsRunnable,
         DEFAULT_MAX_INBOUND_METADATA_SIZE,
-        transportTracer);
+        transportTracer,
+        false);
 
     String host = clientTransport.getOverridenHost();
     int port = clientTransport.getOverridenPort();
@@ -1574,7 +1692,8 @@ public class OkHttpClientTransportTest {
         NO_PROXY,
         tooManyPingsRunnable,
         DEFAULT_MAX_INBOUND_METADATA_SIZE,
-        new TransportTracer());
+        new TransportTracer(),
+        false);
 
     ManagedClientTransport.Listener listener = mock(ManagedClientTransport.Listener.class);
     clientTransport.start(listener);
@@ -1611,7 +1730,8 @@ public class OkHttpClientTransportTest {
             NO_PROXY,
             tooManyPingsRunnable,
             DEFAULT_MAX_INBOUND_METADATA_SIZE,
-            new TransportTracer());
+            new TransportTracer(),
+            false);
 
     ManagedClientTransport.Listener listener = mock(ManagedClientTransport.Listener.class);
     clientTransport.start(listener);
@@ -1643,7 +1763,8 @@ public class OkHttpClientTransportTest {
             .setProxyAddress(serverSocket.getLocalSocketAddress()).build(),
         tooManyPingsRunnable,
         DEFAULT_MAX_INBOUND_METADATA_SIZE,
-        transportTracer);
+        transportTracer,
+        false);
     clientTransport.start(transportListener);
 
     Socket sock = serverSocket.accept();
@@ -1699,7 +1820,8 @@ public class OkHttpClientTransportTest {
             .setProxyAddress(serverSocket.getLocalSocketAddress()).build(),
         tooManyPingsRunnable,
         DEFAULT_MAX_INBOUND_METADATA_SIZE,
-        transportTracer);
+        transportTracer,
+        false);
     clientTransport.start(transportListener);
 
     Socket sock = serverSocket.accept();
@@ -1754,7 +1876,8 @@ public class OkHttpClientTransportTest {
             .setProxyAddress(serverSocket.getLocalSocketAddress()).build(),
         tooManyPingsRunnable,
         DEFAULT_MAX_INBOUND_METADATA_SIZE,
-        transportTracer);
+        transportTracer,
+        false);
     clientTransport.start(transportListener);
 
     Socket sock = serverSocket.accept();
