@@ -23,21 +23,22 @@ import static io.grpc.ConnectivityState.IDLE;
 import static io.grpc.ConnectivityState.READY;
 import static io.grpc.ConnectivityState.SHUTDOWN;
 
+import com.google.common.annotations.VisibleForTesting;
 import com.google.common.base.MoreObjects;
 import com.google.common.base.Objects;
 import com.google.common.base.Stopwatch;
 import com.google.common.base.Supplier;
-import io.grpc.Attributes;
 import io.grpc.CallOptions;
 import io.grpc.ChannelLogger;
 import io.grpc.ChannelLogger.ChannelLogLevel;
 import io.grpc.ClientCall;
 import io.grpc.ConnectivityStateInfo;
-import io.grpc.EquivalentAddressGroup;
 import io.grpc.LoadBalancer;
+import io.grpc.LoadBalancer.CreateSubchannelArgs;
 import io.grpc.LoadBalancer.Factory;
 import io.grpc.LoadBalancer.Helper;
 import io.grpc.LoadBalancer.Subchannel;
+import io.grpc.LoadBalancer.SubchannelStateListener;
 import io.grpc.Metadata;
 import io.grpc.Status;
 import io.grpc.Status.Code;
@@ -52,8 +53,8 @@ import io.grpc.internal.GrpcAttributes;
 import io.grpc.internal.ServiceConfigUtil;
 import io.grpc.util.ForwardingLoadBalancer;
 import io.grpc.util.ForwardingLoadBalancerHelper;
+import io.grpc.util.ForwardingSubchannel;
 import java.util.HashSet;
-import java.util.List;
 import java.util.Map;
 import java.util.concurrent.ScheduledExecutorService;
 import java.util.concurrent.TimeUnit;
@@ -70,8 +71,6 @@ import javax.annotation.Nullable;
  * SynchronizationContext, or it will throw.
  */
 final class HealthCheckingLoadBalancerFactory extends Factory {
-  private static final Attributes.Key<HealthCheckState> KEY_HEALTH_CHECK_STATE =
-      Attributes.Key.create("io.grpc.services.HealthCheckingLoadBalancerFactory.healthCheckState");
   private static final Logger logger =
       Logger.getLogger(HealthCheckingLoadBalancerFactory.class.getName());
 
@@ -91,17 +90,14 @@ final class HealthCheckingLoadBalancerFactory extends Factory {
   public LoadBalancer newLoadBalancer(Helper helper) {
     HelperImpl wrappedHelper = new HelperImpl(helper);
     LoadBalancer delegateBalancer = delegateFactory.newLoadBalancer(wrappedHelper);
-    wrappedHelper.init(delegateBalancer);
     return new HealthCheckingLoadBalancer(wrappedHelper, delegateBalancer);
   }
 
   private final class HelperImpl extends ForwardingLoadBalancerHelper {
     private final Helper delegate;
     private final SynchronizationContext syncContext;
-    
-    private LoadBalancer delegateBalancer;
+
     @Nullable String healthCheckedService;
-    private boolean balancerShutdown;
 
     final HashSet<HealthCheckState> hcStates = new HashSet<>();
 
@@ -110,27 +106,21 @@ final class HealthCheckingLoadBalancerFactory extends Factory {
       this.syncContext = checkNotNull(delegate.getSynchronizationContext(), "syncContext");
     }
 
-    void init(LoadBalancer delegateBalancer) {
-      checkState(this.delegateBalancer == null, "init() already called");
-      this.delegateBalancer = checkNotNull(delegateBalancer, "delegateBalancer");
-    }
-
     @Override
     protected Helper delegate() {
       return delegate;
     }
 
     @Override
-    public Subchannel createSubchannel(List<EquivalentAddressGroup> addrs, Attributes attrs) {
+    public Subchannel createSubchannel(CreateSubchannelArgs args) {
       // HealthCheckState is not thread-safe, we are requiring the original LoadBalancer calls
       // createSubchannel() from the SynchronizationContext.
       syncContext.throwIfNotInThisSynchronizationContext();
+      Subchannel originalSubchannel = super.createSubchannel(args);
       HealthCheckState hcState = new HealthCheckState(
-          this, delegateBalancer, syncContext, delegate.getScheduledExecutorService());
+          this, originalSubchannel, syncContext, delegate.getScheduledExecutorService());
       hcStates.add(hcState);
-      Subchannel subchannel = super.createSubchannel(
-          addrs, attrs.toBuilder().set(KEY_HEALTH_CHECK_STATE, hcState).build());
-      hcState.init(subchannel);
+      Subchannel subchannel = new SubchannelImpl(originalSubchannel, hcState);
       if (healthCheckedService != null) {
         hcState.setServiceName(healthCheckedService);
       }
@@ -150,6 +140,28 @@ final class HealthCheckingLoadBalancerFactory extends Factory {
     }
   }
 
+  @VisibleForTesting
+  static final class SubchannelImpl extends ForwardingSubchannel {
+    final Subchannel delegate;
+    final HealthCheckState hcState;
+
+    SubchannelImpl(Subchannel delegate, HealthCheckState hcState) {
+      this.delegate = checkNotNull(delegate, "delegate");
+      this.hcState = checkNotNull(hcState, "hcState");
+    }
+
+    @Override
+    protected Subchannel delegate() {
+      return delegate;
+    }
+
+    @Override
+    public void start(final SubchannelStateListener listener) {
+      hcState.init(listener);
+      delegate().start(hcState);
+    }
+  }
+
   private static final class HealthCheckingLoadBalancer extends ForwardingLoadBalancer {
     final LoadBalancer delegate;
     final HelperImpl helper;
@@ -169,38 +181,12 @@ final class HealthCheckingLoadBalancerFactory extends Factory {
     }
 
     @Override
-    public void handleResolvedAddressGroups(
-        List<EquivalentAddressGroup> servers, Attributes attributes) {
+    public void handleResolvedAddresses(ResolvedAddresses resolvedAddresses) {
       Map<String, ?> serviceConfig =
-          attributes.get(GrpcAttributes.NAME_RESOLVER_SERVICE_CONFIG);
+          resolvedAddresses.getAttributes().get(GrpcAttributes.NAME_RESOLVER_SERVICE_CONFIG);
       String serviceName = ServiceConfigUtil.getHealthCheckedServiceName(serviceConfig);
       helper.setHealthCheckedService(serviceName);
-      super.handleResolvedAddressGroups(servers, attributes);
-    }
-
-    @Override
-    public void handleSubchannelState(
-        Subchannel subchannel, ConnectivityStateInfo stateInfo) {
-      HealthCheckState hcState =
-          checkNotNull(subchannel.getAttributes().get(KEY_HEALTH_CHECK_STATE), "hcState");
-      hcState.updateRawState(stateInfo);
-
-      if (Objects.equal(stateInfo.getState(), SHUTDOWN)) {
-        helper.hcStates.remove(hcState);
-      }
-    }
-
-    @Override
-    public void shutdown() {
-      super.shutdown();
-      helper.balancerShutdown = true;
-      for (HealthCheckState hcState : helper.hcStates) {
-        // ManagedChannel will stop calling handleSubchannelState() after shutdown() is called,
-        // which is required by LoadBalancer API semantics. We need to deliver the final SHUTDOWN
-        // signal to health checkers so that they can cancel the streams.
-        hcState.updateRawState(ConnectivityStateInfo.forNonError(SHUTDOWN));
-      }
-      helper.hcStates.clear();
+      super.handleResolvedAddresses(resolvedAddresses);
     }
 
     @Override
@@ -211,7 +197,7 @@ final class HealthCheckingLoadBalancerFactory extends Factory {
 
   
   // All methods are run from syncContext
-  private final class HealthCheckState {
+  private final class HealthCheckState implements SubchannelStateListener {
     private final Runnable retryTask = new Runnable() {
         @Override
         public void run() {
@@ -219,13 +205,12 @@ final class HealthCheckingLoadBalancerFactory extends Factory {
         }
       };
 
-    private final LoadBalancer delegate;
     private final SynchronizationContext syncContext;
     private final ScheduledExecutorService timerService;
     private final HelperImpl helperImpl;
-
-    private Subchannel subchannel;
-    private ChannelLogger subchannelLogger;
+    private final Subchannel subchannel;
+    private final ChannelLogger subchannelLogger;
+    private SubchannelStateListener stateListener;
 
     // Set when RPC started. Cleared when the RPC has closed or abandoned.
     @Nullable
@@ -247,18 +232,18 @@ final class HealthCheckingLoadBalancerFactory extends Factory {
 
     HealthCheckState(
         HelperImpl helperImpl,
-        LoadBalancer delegate, SynchronizationContext syncContext,
+        Subchannel subchannel, SynchronizationContext syncContext,
         ScheduledExecutorService timerService) {
       this.helperImpl = checkNotNull(helperImpl, "helperImpl");
-      this.delegate = checkNotNull(delegate, "delegate");
+      this.subchannel = checkNotNull(subchannel, "subchannel");
+      this.subchannelLogger = checkNotNull(subchannel.getChannelLogger(), "subchannelLogger");
       this.syncContext = checkNotNull(syncContext, "syncContext");
       this.timerService = checkNotNull(timerService, "timerService");
     }
 
-    void init(Subchannel subchannel) {
-      checkState(this.subchannel == null, "init() already called");
-      this.subchannel = checkNotNull(subchannel, "subchannel");
-      this.subchannelLogger = checkNotNull(subchannel.getChannelLogger(), "subchannelLogger");
+    void init(SubchannelStateListener listener) {
+      checkState(this.stateListener == null, "init() already called");
+      this.stateListener = checkNotNull(listener, "listener");
     }
 
     void setServiceName(@Nullable String newServiceName) {
@@ -275,13 +260,17 @@ final class HealthCheckingLoadBalancerFactory extends Factory {
       adjustHealthCheck();
     }
 
-    void updateRawState(ConnectivityStateInfo rawState) {
+    @Override
+    public void onSubchannelState(ConnectivityStateInfo rawState) {
       if (Objects.equal(this.rawState.getState(), READY)
           && !Objects.equal(rawState.getState(), READY)) {
         // A connection was lost.  We will reset disabled flag because health check
         // may be available on the new connection.
         disabled = false;
       }
+      if (Objects.equal(rawState.getState(), SHUTDOWN)) {
+        helperImpl.hcStates.remove(this);
+      }
       this.rawState = rawState;
       adjustHealthCheck();
     }
@@ -338,9 +327,9 @@ final class HealthCheckingLoadBalancerFactory extends Factory {
 
     private void gotoState(ConnectivityStateInfo newState) {
       checkState(subchannel != null, "init() not called");
-      if (!helperImpl.balancerShutdown && !Objects.equal(concludedState, newState)) {
+      if (!Objects.equal(concludedState, newState)) {
         concludedState = newState;
-        delegate.handleSubchannelState(subchannel, concludedState);
+        stateListener.onSubchannelState(concludedState);
       }
     }
 
